# -*- coding: utf-8 -*-
# Imported from camel
import re

from nlptools.morphology.charsets import UNICODE_PUNCT_SYMBOL_CHARSET
from nlptools.morphology.charsets import UNICODE_LETTER_MARK_NUMBER_CHARSET


_ALL_PUNCT = u''.join(UNICODE_PUNCT_SYMBOL_CHARSET)
_ALL_LETTER_MARK_NUMBER = u''.join(UNICODE_LETTER_MARK_NUMBER_CHARSET)
_TOKENIZE_RE = re.compile(r'[' + re.escape(_ALL_PUNCT) + r']|[' +
                          re.escape(_ALL_LETTER_MARK_NUMBER) + r']+')


def simple_word_tokenize(sentence):
    """Tokenizes a sentence by splitting on whitespace and seperating
    punctuation. The resulting tokens are either alpha-numeric words or single
    punctuation/symbol characters. This function is language agnostic and
    splits all characters marked as punctuation or symbols in the Unicode
    specification. For example, tokenizing :code:`'Hello,    world!!!'`
    would yield :code:`['Hello', ',', 'world', '!', '!', '!']`.
    Args:
        sentence (:obj:`str`): Sentence to tokenize.
    Returns:
        :obj:`list` of :obj:`str`: The list of tokens.
    """

    return _TOKENIZE_RE.findall(sentence)



 #print(simple_word_tokenize("Hello, world!!!"))



import os
import logging
from collections import namedtuple
from nlptools.arabiner.utils.helpers import load_checkpoint, make_output_dirs, logging_config
from nlptools.arabiner.utils.data import get_dataloaders, parse_conll_files
from nlptools.arabiner.utils.metrics import compute_single_label_metrics, compute_nested_metrics
from nlptools.DataDownload import downloader
logger = logging.getLogger(__name__)


def evaluate_dataset(output_path,data_paths ,batch_size=32):
    """
    Run the model to evaluate text data and save the predictions.

    Args:
        output_path (str): Path to save the results.
        model_path (str): Model path.
        data_paths (list[str]): List of paths to text or sequence files to tag.
        batch_size (int, optional): Batch size. Default is 32.
    """
    # Create directory to save predictions
    make_output_dirs(output_path, overwrite=True)
    logging_config(log_file=os.path.join(output_path, "eval.log"))
    filename = 'Wj27012000.tar'
    path =downloader.get_appdatadir()
    model_path = os.path.join(path, filename)
    # Load tagger
    tagger, tag_vocab, train_config = load_checkpoint(model_path)

    # Convert text to a tagger dataset and index the tokens in args.text
    datasets, vocab = parse_conll_files(data_paths)

    vocabs = namedtuple("Vocab", ["tags", "tokens"])
    vocab = vocabs(tokens=vocab.tokens, tags=tag_vocab)

    # From the datasets generate the dataloaders
    dataloaders = get_dataloaders(
        datasets, vocab,
        train_config.data_config,
        batch_size=batch_size,
        shuffle=[False] * len(datasets)
    )

    # Evaluate the model on each dataloader
    for dataloader, input_file in zip(dataloaders,data_paths):
        filename = os.path.basename(input_file)
        predictions_file = os.path.join(output_path, f"predictions_{filename}")
        _, segments, _, _ = tagger.eval(dataloader)
        tagger.segments_to_file(segments, predictions_file)

        if "Nested" in train_config.trainer_config["fn"]:
            compute_nested_metrics(segments, vocab.tags[1:])
        else:
            compute_single_label_metrics(segments)




import re 

def arStrip(text , diacs=True , smallDiacs=True , shaddah=True , digit=True, alif=True , specialChars=True ):
    
    """
    This method removes Arabic diacritics, small diacritcs, shaddah, Latin and Arabic digits, and unify alif.
    And remove special characters, spaces, underscore and Arabic tatwelah from the input text.

    Args:
        text (:obj:`str`): Arabic text to be processed.
        diacs (:obj:`bool`): flag to remove Arabic diacretics [ ًٌٍَُِْ] (default is True).
        smallDiacs (:obj:`bool`): flag to remove small diacretics (default is True).
        shaddah (:obj:`bool`): flag to remove shaddah (default is True).
        digit (:obj:`bool`): flag to remove Latin and Arabic digits (default is True).
        alif (:obj:`bool`): flag to unify alif (default is True).
        specialChars (:obj:`bool`): flag to remove special characters (default is True).

    Returns:
        :obj:`str`: stripped text.

    **Example:**

    .. highlight:: python
    .. code-block:: python

        from nlptools.utils import parser
        processed_text =parser.arStrip('2023الجو جميلُ')
        print(processed_text)

        #putput
        الجو جميل

        name =parser.arStrip('أَلَمۡ یَأۡنِ لِلَّذِینَ ءَامَنُوۤا۟ أَن تَخۡشَعَ قُلُوبُهُمۡ لِذِكۡرِ ٱللَّهِ وَمَا نَزَلَ مِنَ ٱلۡحَقِّ وَلَا یَكُونُوا۟ كَٱلَّذِینَ أُوتُوا۟ ٱلۡكِتَـٰبَ مِن قَبۡلُ فَطَالَ عَلَیۡهِمُ ٱلۡأَمَدُ فَقَسَتۡ قُلُوبُهُمۡۖ وَكَثِیر مِّنۡهُمۡ فَـسِقُونَ' , True , True , True ,  True , True , True )
        print(name)
        #putput
        الم یان للذین ءامنوا ان تخشع قلوبهم لذكر الله وما نزل من الحق ولا یكونوا كالذین اوتوا الكتٰب من قبل فطال علیهم الامد فقست قلوبهم وكثیر منهم فسقون


    """
    try:
        if text: # if the input string is not empty do the following
            #print("in if")
            if diacs == True :
                text = re.sub(r'[\u064B-\u0650]+', '',text) # Remove all Arabic diacretics [ ًٌٍَُِْ]
                text = re.sub(r'[\u0652]+', '',text) # Remove SUKUN
            if shaddah == True:
                text = re.sub(r'[\u0651]+', '',text) # Remove shddah
            if smallDiacs == True:
                text = re.sub(r'[\u06D6-\u06ED]+', '',text) # Remove all small Quranic annotation signs
            if digit == True:
                text = re.sub('[0-9]+', ' ',text) # Remove English digits
                text = re.sub('[٠-٩]+', ' ',text)# Remove Arabic digits
            
            if alif == True:                             # Unify alif with hamzah: 
                text = re.sub('ٱ', 'ا',text);
                text = re.sub('أ', 'ا',text);
                text = re.sub('إ', 'ا',text);
                text = re.sub('آ', 'ا',text);
            if specialChars == True:
                text = re.sub('[?؟!@#$%-]+' , '' , text) # Remove some of special chars 

            text = re.sub('[\\s]+'," ",text) # Remove all spaces
            text = text.replace("_" , '') #Remove underscore
            text = text.replace("ـ" , '') # Remove Arabic tatwelah
            text = text.strip() # Trim input string
    except:
        return text
    return text
    
def remove_punctuation(text):
    """
    Removes punctuation marks from the text.
    
    Args:
      text (:obj:`str`): The input text.
    
    Returns:
      :obj:`str`: The output text without punctuation marks.

    **Example:**

    .. highlight:: python
    .. code-block:: python
    
        from nlptools.utils import parser
        return parser.remove_punctuation("te!@#،$%%؟st")

        #output
        test

        output= parser.remove_punctuation(" {يَا أَيُّهَا الَّذِينَ آمَنُوا لِيَسْتَأْذِنْكُمُ ....}")
        print(output)

        #output
        يَا أَيُّهَا الَّذِينَ آمَنُوا لِيَسْتَأْذِنْكُمُ  

    """
    try:
        if text:
            punctuation_marks = [r'[\u0021-\u002F]+', r'[U+060C]+', r'[\u003A-\u0040]+',
                                 r'[\u005B-\u0060]+', r'[\u007B-\u007E]+', r'[\u060C]+',
                                 r'[\u061B]+', r'[\u061E]+', r'[\u061F]+', r'[\u0640]+',
                                 r'[\u0653]+', r'[\u065C]+', r'[\u066C]+', r'[\u066A]+',
                                 r'["}"]+', r'["{"]+']
            outputString = text
            for punctuation in punctuation_marks:
                outputString = re.sub(punctuation, '', outputString)
    except:
        return text
    return outputString

def remove_latin(text):
    """
    This method removes all Latin characters from the input text.

    Args:
        text (:obj:`str`): The input text.

    Returns:
         outputString (:obj:`str`): The text without Latin characters.
    Note:
        If an error occurs during processing, the original text is returned.
    **Example:**

    .. highlight:: python
    .. code-block:: python

        from nlptools.utils import parser
        return parser.remove_latin("miojkdujhvaj1546545spkdpoqfoiehwv nWEQFGWERHERTJETAWIKUYFC")
    
        #output
        1546545  

        output =parser.remove_latin("أصل المسمى «تخطيط موارد المؤسسة» هو تعريب لمسمى التخطيط باللغة الإنجليزية Enterprise Resource Planning")
        print(output)

        #output
        أصل المسمى «تخطيط موارد المؤسسة» هو تعريب لمسمى التخطيط باللغة الإنجليزية      
    """
    try:
        if text:
            text = re.sub('[a-zA-Z]+', ' ', text)
    except:
        return text
    return text





from nlptools.salma.wsd import normalizearabert
from nlptools.salma.wsd import GlossPredictor
from nlptools.utils.parser import arStrip
from nlptools.utils.tokenizer import simple_word_tokenize
from nlptools.morphology.alma_multi_word import ALMA_multi_word
from arabiner.bin import infer
from nlptools.morphology.morph_analyzer import analyze
from nlptools.arabiner.bin import infer



def delete_form_list(position, word_lemma):
    tmp_word_lemma = [] 
    output = []
    for wordLemma in word_lemma:
        if position == int(wordLemma[2]): # start 
           word = wordLemma[0]
           gloss = wordLemma[1]
           position = int(wordLemma[3]) 
           concept_count = int(wordLemma[4]) 
           undiac_multi_word_lemma = wordLemma[5]
           multi_word_lemma = wordLemma[6]
           output.append([word, gloss, concept_count, undiac_multi_word_lemma, multi_word_lemma])# word
        elif position < int(wordLemma[2]): 
           tmp_word_lemma.append(wordLemma)
    return tmp_word_lemma, output, position
    #return output

def find_two_word_lemma(input_sentence):
    i = 0
    output = []
    length = len(input_sentence)
    while i < length - 1:
        two_grams = input_sentence[i] +" "+ input_sentence[i + 1] 
        data = ALMA_multi_word(two_grams,2)
        try :
            # found two_grams
            found_2Word_lemma = [two_grams,data[0]['glosses'], i, i + 1,data[0]['concept_count'], data[0]['undiac_multi_word_lemma'], data[0]['multi_word_lemma']]
            output.append(found_2Word_lemma) 
            i = i + 1    
        except: # no record found on this multi_lema
            i = i + 1 
    return output


def find_three_word_lemma(input_sentence):
    i = 0
    output = []
    length = len(input_sentence)
    while i < length - 2:
        three_grams = input_sentence[i] +" "+ input_sentence[i + 1] + " "+ input_sentence[i + 2]
        data = ALMA_multi_word(three_grams,3)
        try:
           found_3Word_lemma = [three_grams, data[0]['glosses'], i, i + 2,data[0]['concept_count'], data[0]['undiac_multi_word_lemma'], data[0]['multi_word_lemma']]
           output.append(found_3Word_lemma) 
           i = i + 1    
        except:  
           i = i + 1 
    return output

def find_four_word_lemma(input_sentence):
   i = 0
   output = []
   length = len(input_sentence)
   while i < length - 3:
      four_grams = input_sentence[i] +" "+ input_sentence[i + 1] + " "+ input_sentence[i + 2] + " "+ input_sentence[i + 3]
      data = ALMA_multi_word(four_grams,4)
      try:
         found_4Word_lemma = [four_grams, data[0]['glosses'], i, i + 3,data[0]['concept_count'], data[0]['undiac_multi_word_lemma'], data[0]['multi_word_lemma']]
         output.append(found_4Word_lemma) 
         i = i + 1    
      except:  
         i = i + 1 
   return output


def find_five_word_lemma(input_sentence):
   i = 0
   output = []
   length = len(input_sentence)
   while i < length - 4:
      five_grams = input_sentence[i] +" "+ input_sentence[i + 1] + " "+ input_sentence[i + 2] + " "+ input_sentence[i + 3] + " "+ input_sentence[i + 4]
      data = ALMA_multi_word(five_grams,5)
      try:
         found_5Word_lemma = [five_grams, data[0]['glosses'], i, i + 4,data[0]['concept_count'], data[0]['undiac_multi_word_lemma'], data[0]['multi_word_lemma']]
         output.append(found_5Word_lemma) 
         i = i + 1    
      except:  
         i = i + 1 
   return output

def find_named_entities(string):
   found_entities = []
   entites = infer.ner(string, "4")
   tag_gloss = {
      "PERS": "اسم شخص",
      "ORG": "اسم مؤسسة",
      "LOC": "اسم منطقة جغرافية",
      "FAC": "اسم لمَعلَم",
      "DATE": "فترة زمنية تدل على تاريخ",
      "UNIT": "وحدة قياس",
      "CURR": "عملة",
      "GPE": "اسم بلد، له حدود إدارية/جيوسياسية",
      "TIME": "فترة زمنية تدل على الوقت",
      "CARDINAL": "عدد يدل على معدود",
      "ORDINAL": "رقم، لا يدل على معدود",
      "PERCENT": "نسبة مئوية",
      "QUANTITY": "كمية",
      "MONEY": "مبلغ مالي",
      "LANGUAGE": "اسم للغة طبيعية",
      "PRODUCT": "اسم منتج",
      "LAW": "قانون"
   }

   for entity in entites:
      gloss_ner = ""
      if entity[1] in tag_gloss.keys():
         gloss_ner = tag_gloss[entity[1]]  

      if gloss_ner != "":
         gloss = [{'concept_id': '', 'resource_id': '', 'resource_name': '', 'gloss': gloss_ner}]   
         entity = [entity[0],gloss,int(entity[2]), int(entity[3]),1,arStrip(entity[0],True,True,True,False,True,False),entity[0]]   
         found_entities.append(entity)

   return found_entities   


def find_glosses_using_ALMA(word):

   data = analyze(word)
   Diac_lemma = ""
   pos = ""
   Undiac_lemma = ""
   glosses = []

   Diac_lemma = data[0][1]
   pos = data[0][2]
   Undiac_lemma = arStrip(Diac_lemma, True, True, True, True, True, False) # Remove diacs , smallDiacs , shaddah ,  digit , alif , specialChars
   glosses = data[0][4]
   concept_count = data[0][3]
   return word, Undiac_lemma, Diac_lemma, pos , concept_count, glosses
   
def disambiguate_glosses_using_SALMA(glosses, Diac_lemma, Undiac_lemma, word, sentence):
   word = normalizearabert(word)

   glosses_dictionary = {}
   if glosses != None:
      for gloss in glosses:

         glosses_dictionary.update({gloss['concept_id'] : gloss['gloss']}) 

      concept_id, gloss = GlossPredictor(Diac_lemma, Undiac_lemma,word,sentence,glosses_dictionary)
      my_json = {}    
      my_json['Concept_id'] = concept_id
      my_json['Gloss'] = gloss
      my_json['word'] = word
      my_json['Undiac_lemma'] = Undiac_lemma
      my_json['Diac_lemma'] = Diac_lemma
      return my_json
   else:

      my_json = {}    
      my_json['word'] = word
      my_json['Undiac_lemma'] = Undiac_lemma
      my_json['Diac_lemma'] = Diac_lemma
      return my_json




def find_glosses(input_sentence, three_word_lemma, two_word_lemma, four_word_lemma, five_word_lemma, ner):
      output_list = []
      position = 0
      while position < len(input_sentence):    
         flag = "False"
         output_from5word = delete_form_list(position, five_word_lemma)
         five_word_lemma = output_from5word[0]
         if output_from5word[1] != []: # output
            position = output_from5word[2]  
            flag = "True"
            my_json = {}    
            my_json['word'] = output_from5word[1][0][0]
            my_json['concept_count'] = output_from5word[1][0][2]
            my_json['glosses'] = output_from5word[1][0][1]
            my_json['Diac_lemma'] = output_from5word[1][0][4]
            my_json['Undiac_lemma'] = output_from5word[1][0][3]
            output_list.append(my_json)
            position = position + 1                



         output_from4word = delete_form_list(position, four_word_lemma)
         four_word_lemma = output_from4word[0]
         if output_from4word[1] != []: # output
            position = output_from4word[2]  
            flag = "True"
            my_json = {}    
            my_json['word'] = output_from4word[1][0][0]
            my_json['concept_count'] = output_from4word[1][0][2]
            my_json['glosses'] = output_from4word[1][0][1]
            my_json['Diac_lemma'] = output_from4word[1][0][4]
            my_json['Undiac_lemma'] = output_from4word[1][0][3]
            output_list.append(my_json)
            position = position + 1                
         
         output_from3word = delete_form_list(position, three_word_lemma)
         three_word_lemma = output_from3word[0]
         if output_from3word[1] != []: # output
            position = output_from3word[2]  
            flag = "True"
            my_json = {}    
            my_json['word'] = output_from3word[1][0][0]
            my_json['concept_count'] = output_from3word[1][0][2]
            my_json['glosses'] = output_from3word[1][0][1]
            my_json['Diac_lemma'] = output_from3word[1][0][4]
            my_json['Undiac_lemma'] = output_from3word[1][0][3]
            output_list.append(my_json)
            position = position + 1                



         output_from2Word = delete_form_list(position, two_word_lemma)
         two_word_lemma = output_from2Word[0] 
         if output_from2Word[1] != []:  
            position = output_from2Word[2]
            flag = "True"
            my_json = {}    
            word = output_from2Word[1][0][0]
            my_json['word'] = word
            my_json['concept_count'] = output_from2Word[1][0][2]
            my_json['glosses'] = output_from2Word[1][0][1]
            my_json['Diac_lemma'] = output_from2Word[1][0][4]
            my_json['Undiac_lemma'] = output_from2Word[1][0][3] 
            output_list.append(my_json)
            position = position + 1                 
               


         output_from_ner = delete_form_list(position, ner)
         ner = output_from_ner[0] 
         if output_from_ner[1] != []:  
            position = output_from_ner[2]
            flag = "True"
            my_json = {}    
            word = output_from_ner[1][0][0]
            my_json['word'] = word
            my_json['concept_count'] = output_from_ner[1][0][2]
            my_json['glosses'] = output_from_ner[1][0][1]
            my_json['Diac_lemma'] = output_from_ner[1][0][4]
            my_json['Undiac_lemma'] = output_from_ner[1][0][3] 
            output_list.append(my_json)
            position = position + 1                             
         
         if flag == "False": # Not found in ner or in multi_word_dictionary, ASK ALMA 
            word = input_sentence[position]
            word, Undiac_lemma, Diac_lemma, pos , concept_count, glosses = find_glosses_using_ALMA(word)   
            my_json = {}    
            my_json['word'] = word
            my_json['concept_count'] = concept_count
            my_json['glosses'] = glosses
            my_json['Diac_lemma'] = Diac_lemma
            my_json['Undiac_lemma'] = Undiac_lemma
            output_list.append(my_json)
            position = position + 1  
      return output_list                    

def disambiguate_glosses_main(word, sentence):
   concept_count = word['concept_count']
   if concept_count == 0:
      print(" word : ", word['word'], " with concept count = 0", concept_count)
      my_json = {}    
      my_json['word'] = word['word']
      my_json['Diac_lemma'] = word['Diac_lemma']
      my_json['Undiac_lemma'] = word['Undiac_lemma']
      return my_json
   elif concept_count == 1:
      print(" word : ", word['word'], " with concept count = 1", concept_count)
      my_json = {}    
      my_json['word'] = word['word']
      glosses = word['glosses'][0]
      my_json['Gloss'] = glosses['gloss']
      my_json['Concept_id'] = glosses['concept_id']
      my_json['Diac_lemma'] = word['Diac_lemma']
      my_json['Undiac_lemma'] = word['Undiac_lemma']
      return my_json
   else:   
      print(" word : ", word['word'], " with concept count > 1", concept_count)
      input_word = word['word']
      concept_count = word['concept_count']
      glosses = word['glosses']
      Diac_lemma = word['Diac_lemma']
      Undiac_lemma = word['Undiac_lemma']
      return disambiguate_glosses_using_SALMA(glosses, Diac_lemma, Undiac_lemma, input_word, sentence)

def WSD(sentence):
   
   input_sentence = simple_word_tokenize(sentence)
   
   five_word_lemma = find_five_word_lemma(input_sentence)
   
   four_word_lemma = find_four_word_lemma(input_sentence)
   
   three_word_lemma = find_three_word_lemma(input_sentence)
   
   two_word_lemma = find_two_word_lemma(input_sentence)
   
   ner = find_named_entities(" ".join(input_sentence))

   output_list = find_glosses(input_sentence, three_word_lemma, two_word_lemma, four_word_lemma, five_word_lemma, ner)
   
   results = []

   for word in output_list:
      results.append(disambiguate_glosses_main(word, sentence))
   return results


def SALMA(sentence):
      results=''
      if len(sentence) > 500:        
         results = "Input is too long"
      else: 
         results = WSD(sentence)
      return results
              


from nlptools.morphology import settings 
import pickle
from nlptools.DataDownload import downloader
import os 


filename = 'ALMA27012000.pickle'
path =downloader.get_appdatadir()
file_path = os.path.join(path, filename)
with open(file_path, 'rb') as f:
       #Load the serialized data from the file
    settings.div_dic = pickle.load(f)
       #print(ALMA_dic)

filename_five = 'five_grams.pickle'
path =downloader.get_appdatadir()
file_path = os.path.join(path, filename_five)
with open(file_path, 'rb') as f:
       #Load the serialized data from the file
    settings.five_grams_dict = pickle.load(f, encoding='utf-8')
       #print(ALMA_dic)



filename_four = 'four_grams.pickle'
path =downloader.get_appdatadir()
file_path = os.path.join(path, filename_four)
with open(file_path, 'rb') as f:
       #Load the serialized data from the file
    settings.four_grams_dict = pickle.load(f, encoding='utf-8')
       #print(ALMA_dic)
       
filename_three = 'three_grams.pickle'
path =downloader.get_appdatadir()
file_path = os.path.join(path, filename_three)
with open(file_path, 'rb') as f:
       #Load the serialized data from the file
    settings.three_grams_dict = pickle.load(f, encoding='utf-8')
       #print(ALMA_dic)



filename_two = 'two_grams.pickle'
path =downloader.get_appdatadir()
file_path = os.path.join(path, filename_two)
with open(file_path, 'rb') as f:
       #Load the serialized data from the file
    settings.two_grams_dict = pickle.load(f, encoding='utf-8')
       #print(ALMA_dic)


# Created: 07/03/2021 , By Mohammad A'bed , Tested on SAMA lemma DB  
# Trello task: https://trello.com/c/wsiMd4iW/20-buckwalter-to-arabic-and-vice-versa
from collections import deque

# This is a mapping dictionery of Arabic letters to BW 
# dictionery contains: key -> Unicode to Arabic, value -> BW chars
# It includes all BW mappings in addition to other special characters
#  that are used in the SAMA database but not part of the BW character set such as numbers and Qur'anic diacritics
ar2bw_map = {
    '\u0621' : '\'' , # ء
    '\u0622' : '|' , # آ
    '\u0623' : '>' , # أ
    '\u0624' : '&' , # ؤ
    '\u0625' : '<' , # إ
    '\u0626' : '}' , # ئ
    '\u0627' : 'A' , # ا
    '\u0628' : 'b' , # ب
    '\u0629' : 'p' , # ة
    '\u062A' : 't' , # ت
    '\u062B' : 'v' , # ث
    '\u062C' : 'j' , # ج
    '\u062D' : 'H' , # ح
    '\u062E' : 'x' , # خ
    '\u062F' : 'd' , # د
    '\u0630' : '*' , # ذ
    '\u0631' : 'r' , # ر
    '\u0632' : 'z' , # ز
    '\u0633' : 's' , # س
    '\u0634' : '$' , # ش
    '\u0635' : 'S' , # ص
    '\u0636' : 'D' , # ض
    '\u0637' : 'T' , # ط
    '\u0638' : 'Z' , # ظ
    '\u0639' : 'E' , # ع
    '\u063A' : 'g' , # غ
    '\u0020' : ' ' , #Space is replaced by space 
    '\u0640' : '_' , #ـ
    '\u0641' : 'f' , # ف
    '\u0642' : 'q' , # ق
    '\u0643' : 'k' , # ك
    '\u0644' : 'l' , # ل
    '\u0645' : 'm' , # م
    '\u0646' : 'n' , # ن
    '\u0647' : 'h' , # ه
    '\u0648' : 'w' , # و
    '\u0649' : 'Y' , # ى
    '\u064A' : 'y' , # ي
    '\u064B' : 'F' , # TANWEEN FATH ً
    '\u064C' : 'N' , #TANWEEN DHAM ٌ
    '\u064D' : 'K' , #TANWEE KASR ٍ
    '\u064E' : 'a' , #FATHA َ
    '\u064F' : 'u' , #DHAMMA ُ
    '\u0650' : 'i' , #KASRA ِ
    '\u0651' : '~' , # SHADDAH ّ
    '\u0652' : 'o' , #SUKON ْ
    '\u0670' : '`' , # SHORT ALEF (dagger alif) ٰ
    '\u0671' : '{' , #AL WITH HAMZA QATE'/WASEL MARK ٱ
    '\u067E' : 'P' , # پ 
    '\u0686' : 'J' , # چ 
    '\u06A4' : 'V' , # ڤ 
    '\u06AF' : 'G' , # گ 
    '\u06EA' : '-' , # ۪ Not BW standerd
    '\u0654' : '#' , # ٔ     Not BW standerd
    '\u06DC ' : ':' , # short sen ۜ  Not BW standerd  
    '\u06E0' : '"' , # small circle ۟  ,Not BW standerd
    '\u06E2' : '[' , # short meem ۢ  , Not BW standerd
    '\u06E3 ' : ';' , #small sen below letter  ۣ  , Not BW standerd
    '\u06E5 ' : ',' , # samall wa ۥ  , Not BW standerd
    '\u06E6' : '.' , # ۦ   Not BW standerd
    '\u06E8 ' : '!' , # small noon ۨ  Not BW standerd
    '\u06EA' : '-' , # special qura'nic arabic ( littel circule underneath), is not part of standerd BW , not in Buckwalter ۪  Not BW standerd
    '\u06EC' : '%' , # small solid circle ۬  Not BW standerd
    '\u06ED ' : ']',  # small meem ۭ  Not BW standerd
    '0' : '0' , #
    '1' : '1' , #
    '2' : '2' , #
    '3' : '3' , #
    '4' : '4' , #
    '5' : '5' , #
    '6' : '6' , #
    '7' : '7' , #
    '8' : '8' , #
    '9' : '9'  #
}

# This is a mapping dictionery of BW  letters to  Arabic
# It includes all Arabic mappings in addition to other special characters
#  that are used in the SAMA database but not part of the Arabic character set such as numbers and Qur'anic diacritics
bw2ar_map = {
    '\'' : '\u0621' , # ء
    '|' : '\u0622' , # آ
    '>' : '\u0623' , # أ
    '&' : '\u0624' , # ؤ
    '<' : '\u0625' , # إ
    '}' : '\u0626' , # ئ
    'A' : '\u0627' , # ا
    'b' : '\u0628' , # ب
    'p' : '\u0629' , # ة
    't' : '\u062A' , # ت
    'v' : '\u062B' , # ث
    'j' : '\u062C' , # ج
    'H' : '\u062D' , # ح
    'x' : '\u062E' , # خ
    'd' : '\u062F' , # د
    '*' : '\u0630' , # ذ
    'r' : '\u0631' , # ر
    'z' : '\u0632' , # ز
    's' : '\u0633' , # س
    '$' : '\u0634' , # ش
    'S' : '\u0635' , # ص
    'D' : '\u0636' , # ض
    'T' : '\u0637' , # ط
    'Z' : '\u0638' , # ظ
    'E' : '\u0639' , # ع
    'g' : '\u063A' , # غ
    ' ' : '\u0020' , #Space is replaced by space 
    '_' : ' ' , #ـ    This is Temporary
    # '_' : '\u0640' , #ـ
    'f' : '\u0641' , # ف
    'q' : '\u0642' , # ق
    'k' : '\u0643' , # ك
    'l' : '\u0644' , # ل
    'm' : '\u0645' , # م
    'n' : '\u0646' , # ن
    'h' : '\u0647' , # ه
    'w' : '\u0648' , # و
    'Y' : '\u0649' , # ى
    'y' : '\u064A' , # ي
    'F' : '\u064B' , # TANWEEN FATH  ً
    'N' : '\u064C' , #TANWEEN DHAM  ٌ
    'K' : '\u064D' , #TANWEE KASR  ٍ
    'a' : '\u064E' , #FATHA  َ
    'u' : '\u064F' , #DHAMMA  ُ
    'i' : '\u0650' , #KASRA  ِ
    '~' : '\u0651' , # SHADDAH  ّ
    'o' : '\u0652' , #SUKON  ْ
    '`' : '\u0670' , # SHORT ALEF (dagger alif) ٰ
    '{' : '\u0671' , #AL WITH HAMZA QATE'/WASEL MARK ٱ
    'P' : '\u067E' , # پ 
    'J' : '\u0686' , # چ 
    'V' : '\u06A4' , # ڤ 
    'G' : '\u06AF' , # گ 
    '-' : '\u06EA' , # ۪ Not BW standerd
    '#' : '\u0654' , # ٔ  ARABIC HAMZA ABOVE   Not BW standerd
    ':' : '\u06DC ' , # short sen ۜ  Not BW standerd  
    '"' : '\u06E0' , # small circle ۟  ,Not BW standerd
    '[' : '\u06E2' , # short meem ۢ  , Not BW standerd
    ';' : '\u06E3 ' , #small sen below letter  ۣ  , Not BW standerd
    ',' : '\u06E5 ' , # samall wa ۥ  , Not BW standerd
    '.' : '\u06E6' , # ۦ   Not BW standerd
    '!' : '\u06E8 ' , # small noon ۨ  Not BW standerd
    '-' : '\u06EA' , # special qura'nic arabic ( littel circule underneath), is not part of standerd BW , not in Buckwalter ۪  Not BW standerd
    '%' : '\u06EC' , # small solid circle ۬  Not BW standerd
    ']' : '\u06ED ',  # small meem ۭ  Not BW standerd
    '0' : '0' , #
    '1' : '1' , #
    '2' : '2' , #
    '3' : '3' , #
    '4' : '4' , #
    '5' : '5' , #
    '6' : '6' , #
    '7' : '7' , #
    '8' : '8' , #
    '9' : '9'  #
}
# A transliterate Function to transliterate Arabic letters and vice versa
#It takes a text and the schema as input and return 2-values: the transliteration and a flag of whether all chars are transliterated or not
def perform_transliteration(text , schema ):
    """
    This method takes a text and a schema as input and returns a tuple of two values: the transliteration of the text based on the given schema and a flag indicating whether all characters in the text were transliterated or not.

    Args:
        text (:obj:`str`): The input text to be transliterated.
        schema (:obj:`str`): The transliteration schema to be used. Should be either `bw2ar` to transliterate Buckwalter-encoded Arabic text to Arabic script, or `ar2bw` to transliterate Arabic script to Buckwalter-encoded Arabic text.

    Returns:
        :obj:`tuple`: A tuple of two values:
        - The transliterated text based on the given schema.
        - A boolean flag indicating whether all characters in the input text were successfully transliterated or not.

    **Example:**

    .. highlight:: python
    .. code-block:: python

        from nlptools.utils import text_transliteration

        print(text_transliteration.perform_transliteration("مُحَمَدٌ نَـشِيْطٌـ1"  , "ar2bw"))
        print(text_transliteration.perform_transliteration("muHamadN"  , "bw2ar"))
        print(text_transliteration.perform_transliteration("شَنُعُ۪ـ1"  , "ar2bw"))
        print(text_transliteration.perform_transliteration("$aw~aH_2"  , "bw2ar"))

        #output
        ('muHamadN na_$iyoTN_1', True)
        ('مُحَمَدٌ', True)
        ('$anuE-u_1', True)
        ('شَوَّح 2', True)
    """
    # to map BW into Arabic  
    if schema == "bw2ar":
        transliterated_text = deque() #Empty deque list
        
        is_all_mapped = True
        for char in text:
            # lockup every charecters in the dictionary, if not found return the original char 
            char_value = bw2ar_map.get(char)
   
            if char_value is None:
                is_all_mapped = False  # False if one cjars is not in map
                transliterated_text.append(char)
            else:
                transliterated_text.append(char_value)

        return ''.join(transliterated_text)  , is_all_mapped
    # to map Arabic into BW
    elif schema == "ar2bw":

        transliterated_text = deque()
   
        is_all_mapped = True
        for char in text:
            # lockup evry charecters in the dictionary, if not found return the original char 
            char_value = ar2bw_map.get(char)
          
            if char_value is None:
                is_all_mapped = False   # False if one cjars is not in map
                transliterated_text.append(char)
            else:
                transliterated_text.append(char_value)

    return ''.join(transliterated_text)  , is_all_mapped






from torch.utils.data import DataLoader
from torchtext.vocab import vocab
from collections import Counter, namedtuple
import logging
import re
import itertools
from nlptools.arabiner.utils.helpers import load_object
from nlptools.arabiner.data.datasets import Token

logger = logging.getLogger(__name__)


def conll_to_segments(filename):
    """
    Convert CoNLL files to segments. This return list of segments and each segment is
    a list of tuples (token, tag)
    :param filename: Path
    :return: list[[tuple]] - [[(token, tag), (token, tag), ...], [(token, tag), ...]]
    """
    segments, segment = list(), list()

    with open(filename, "r") as fh:
        for token in fh.read().splitlines():
            if not token.strip():
                segments.append(segment)
                segment = list()
            else:
                parts = token.split()
                token = Token(text=parts[0], gold_tag=parts[1:])
                segment.append(token)

        segments.append(segment)

    return segments


def parse_conll_files(data_paths):
    """
    Parse CoNLL formatted files and return list of segments for each file and index
    the vocabs and tags across all data_paths
    :param data_paths: tuple(Path) - tuple of filenames
    :return: tuple( [[(token, tag), ...], [(token, tag), ...]], -> segments for data_paths[i]
                    [[(token, tag), ...], [(token, tag), ...]], -> segments for data_paths[i+1],
                    ...
                  )
             List of segments for each dataset and each segment has list of (tokens, tags)
    """
    vocabs = namedtuple("Vocab", ["tags", "tokens"])
    datasets, tags, tokens = list(), list(), list()

    for data_path in data_paths:
        dataset = conll_to_segments(data_path)
        datasets.append(dataset)
        tokens += [token.text for segment in dataset for token in segment]
        tags += [token.gold_tag for segment in dataset for token in segment]

    # Flatten list of tags
    tags = list(itertools.chain(*tags))

    # Generate vocabs for tags and tokens
    tag_vocabs = tag_vocab_by_type(tags)
    tag_vocabs.insert(0, vocab(Counter(tags)))
    vocabs = vocabs(tokens=vocab(Counter(tokens), specials=["UNK"]), tags=tag_vocabs)
    return tuple(datasets), vocabs


def tag_vocab_by_type(tags):
    vocabs = list()
    c = Counter(tags)
    tag_names = c.keys()
    tag_types = sorted(list(set([tag.split("-", 1)[1] for tag in tag_names if "-" in tag])))

    for tag_type in tag_types:
        r = re.compile(".*-" + tag_type)
        t = list(filter(r.match, tags)) + ["O"]
        vocabs.append(vocab(Counter(t), specials=["<pad>"]))

    return vocabs


def text2segments(text):
    """
    Convert text to a datasets and index the tokens
    """
    dataset = [[Token(text=token, gold_tag=["O"]) for token in text.split()]]
    tokens = [token.text for segment in dataset for token in segment]

    # Generate vocabs for the tokens
    segment_vocab = vocab(Counter(tokens), specials=["UNK"])
    return dataset, segment_vocab


def get_dataloaders(
    datasets, vocab, data_config, batch_size=32, num_workers=0, shuffle=(True, False, False)
):
    """
    From the datasets generate the dataloaders
    :param datasets: list - list of the datasets, list of list of segments and tokens
    :param batch_size: int
    :param num_workers: int
    :param shuffle: boolean - to shuffle the data or not
    :return: List[torch.utils.data.DataLoader]
    """
    dataloaders = list()

    for i, examples in enumerate(datasets):
        data_config["kwargs"].update({"examples": examples, "vocab": vocab})
        dataset = load_object(data_config["fn"], data_config["kwargs"])

        dataloader = DataLoader(
            dataset=dataset,
            shuffle=shuffle[i],
            batch_size=batch_size,
            num_workers=num_workers,
            collate_fn=dataset.collate_fn,
        )

        logger.info("%s batches found", len(dataloader))
        dataloaders.append(dataloader)

    return dataloaders



from nlptools.morphology import settings 
from nlptools.utils.parser import arStrip
import json

def ALMA_multi_word(multi_word, n_gram):
    undiac_multi_word = arStrip(multi_word, True, True, True, False, True, False)  # diacs , smallDiacs , shaddah ,  digit , alif , specialChars
    #print(settings.five_grams_dict[undiac_multi_word])
    result_word = []
    if n_gram == 2:
        if undiac_multi_word in settings.two_grams_dict.keys():
           result_word = settings.two_grams_dict[undiac_multi_word]
       #result_word = json.loads(serializers.serialize('json',models.multi_word_lemmas.objects.filter(n_gram=2).filter(undiac_multi_word_lemma=undiac_multi_word)))
    elif n_gram == 3:    
        if undiac_multi_word in settings.three_grams_dict.keys():
           result_word = settings.three_grams_dict[undiac_multi_word]
       #result_word = json.loads(serializers.serialize('json',models.multi_word_lemmas.objects.filter(n_gram=3).filter(undiac_multi_word_lemma=undiac_multi_word)))
    elif n_gram == 4:    
        if undiac_multi_word in settings.four_grams_dict.keys():
           result_word = settings.four_grams_dict[undiac_multi_word]
       #result_word = json.loads(serializers.serialize('json',models.multi_word_lemmas.objects.filter(n_gram=4).filter(undiac_multi_word_lemma=undiac_multi_word))) 
    elif n_gram == 5:
        if undiac_multi_word in settings.five_grams_dict.keys():
           result_word = settings.five_grams_dict[undiac_multi_word]
       #result_word = json.loads(serializers.serialize('json',models.multi_word_lemmas.objects.filter(n_gram=5).filter(undiac_multi_word_lemma=undiac_multi_word)))
    

    my_json = {}
    glosses_list = []
    output_list = []
    concept_count = 0
    my_json['multi_word_lemma'] = multi_word
    my_json['undiac_multi_word_lemma'] = multi_word
    if result_word != [] :
       my_json['multi_word_lemma'] = result_word[0]  # multi_word_lemma
       #my_json['concept_count'] = result_word[1] #concept_count
       my_json['POS'] = result_word[2] #POS

       glosses_list.append(json.loads(result_word[3]))
       concept_count = concept_count + 1
       
       my_json['concept_count'] = concept_count
       my_json['glosses'] = glosses_list   
       output_list.append(my_json)    
    return output_list 